SystemOrganization addCategory: #MetaInterpreter!
SystemOrganization addCategory: #'MetaInterpreter-CaseStudy'!

TestCase subclass: #TestInterpreter
	instanceVariableNames: 'anIv'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/8/2013 16:13'!
passAsArg: t

! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/17/2013 11:39'!
store: t

	anIv := t

! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 9/3/2013 11:46'!
testDerivativeHooksWithHalt

	"press proceed for the test to finish"

	|t1 t2|
	
	"register halt on creation"
	Interpreter on: MessageReceived for: TestObject do: [:reifications :reflectogram | 
		reifications message selector = #new ifTrue: [self halt].
	].

	Interpreter enableMetaBehaviorFor: TestObject.
	
	t1 := TestObject new. "should halt"
	t2 := TestObject new. "should halt."
	
	Interpreter removeMetaBehaviorOf: TestObject.
	
	"register halt on object in invoke -- this can be done either by registering the argument or the receiver"
	Interpreter on: AsArgument for: t2 do: [:reifications :reflectogram |
			reifications receiver = t1 ifTrue: [self halt].
	].

	Interpreter enableMetaBehaviorFor: t2.
	
	t1 testPassingArgument: t2. "should halt"

	Interpreter removeMetaBehaviorOf: t2.
	
	"register halt on interaction"	
	Interpreter on: MessageReceived for: t1 do: [:reifications :reflectogram | 
		(reifications message arguments includes: t2) ifTrue: [self halt].
	].

	Interpreter on: AsArgument for: t1 do: [:reifications :reflectogram | 
		reifications receiver = t2 ifTrue: [self halt].
	].

	Interpreter enableMetaBehaviorFor: t1.


	t1 testInteraction: t2. "should halt because t1 invokes a method of t2"
	t2 testInteraction: t1. "should halt because t2 invokes a method of t1"
	
	Interpreter disableMetaBehaviorFor: t1.
! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/4/2013 15:45'!
testHookArg

	|t receiveCounter sendCounter ivReadCounter ivWriteCounter argCounter|
	
	receiveCounter := 0.
	sendCounter := 0.
	ivReadCounter := 0.
	ivWriteCounter := 0.
	argCounter := 0.

	1 to: 100 do: [:i| 	
		
		t := TestObject new.
	
		Interpreter on: AsArgument for: t do: [:reifications :reflectogram |
			argCounter := argCounter + 1.
		].

		Interpreter enableMetaBehaviorFor: t.
	
		self passAsArg: t. "should trigger meta arg"
		t x: 5. "should trigger meta receive and meta-write".
		self assert: (t x = 5). "should trigger meta-receive and meta-read".
		t testSenderHook. "should trigger 2 meta-sends"
	
		"Interpreter disableMetaBehaviorFor: t."
		Interpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	].
	
	self assert: argCounter = (1*100).! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 3/13/2014 09:00'!
testHookInvoke

	|t invokeCounter |
	
	invokeCounter := 0.

	1 to: 100 do: [:i| 	
		
		t := TestObject new.
	
	
		Interpreter on: MethodInvoke for: t class>>#testSenderHook do: [:reifications :reflectogram | 
			invokeCounter := invokeCounter + 1.	
		].
	

		Interpreter enableMetaBehaviorFor: t class>>#testSenderHook.
	
		t testSenderHook. "should trigger 2 meta-sends"
	
		"Interpreter disableMetaBehaviorFor: t."
		Interpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	].
	
	self assert: invokeCounter = (1*100).
! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 3/14/2014 06:12'!
testHookMethodSend

	|t invokeCounter |
	
	invokeCounter := 0.

	1 to: 100 do: [:i| 	
		
		t := TestObject new.
	
	
		Interpreter on: MethodSend for: t class>>#testSenderHook do: [:reifications :reflectogram | 
			invokeCounter := invokeCounter + 1.	
		].
	

		Interpreter enableMetaBehaviorFor: t class>>#testSenderHook.
	
		t testSenderHook. "should trigger 2 meta-sends"
	
		"Interpreter disableMetaBehaviorFor: t."
		Interpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	].
	
	self assert: invokeCounter = (2*100).
! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/4/2013 15:45'!
testHookRead

	|t receiveCounter sendCounter ivReadCounter ivWriteCounter argCounter|
	
	receiveCounter := 0.
	sendCounter := 0.
	ivReadCounter := 0.
	ivWriteCounter := 0.
	argCounter := 0.

	1 to: 100 do: [:i| 	
		
		t := TestObject new.
	
		Interpreter on: IvRead for: t do: [:reifications :reflectogram | 
			ivReadCounter := ivReadCounter + 1.
		].

		Interpreter enableMetaBehaviorFor: t.
	
		self passAsArg: t. "should trigger meta arg"
		t x: 5. "should trigger meta receive and meta-write".
		self assert: (t x = 5). "should trigger meta-receive and meta-read".
		t testSenderHook. "should trigger 2 meta-sends"
	
		"Interpreter disableMetaBehaviorFor: t."
		Interpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	].
	
	self assert: ivReadCounter = (1*100).
! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/4/2013 15:46'!
testHookReceive

	|t receiveCounter sendCounter ivReadCounter ivWriteCounter argCounter|
	
	receiveCounter := 0.
	sendCounter := 0.
	ivReadCounter := 0.
	ivWriteCounter := 0.
	argCounter := 0.

	1 to: 100 do: [:i| 	
		
		t := TestObject new.
	
		Interpreter on: MessageReceived for: t do: [:reifications :reflectogram | 
			receiveCounter := receiveCounter + 1.
		].

		Interpreter enableMetaBehaviorFor: t.
	
		self passAsArg: t. "should trigger meta arg"
		t x: 5. "should trigger meta receive and meta-write".
		self assert: (t x = 5). "should trigger meta-receive and meta-read".
		t testSenderHook. "should trigger 2 meta-sends"
	
		"Interpreter disableMetaBehaviorFor: t."
		Interpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	].
	
	self assert: receiveCounter = (3*100).
! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/4/2013 15:46'!
testHookSend

	|t receiveCounter sendCounter ivReadCounter ivWriteCounter argCounter|
	
	receiveCounter := 0.
	sendCounter := 0.
	ivReadCounter := 0.
	ivWriteCounter := 0.
	argCounter := 0.

	1 to: 100 do: [:i| 	
		
		t := TestObject new.
	
	
		Interpreter on: MessageSent for: t do: [:reifications :reflectogram | 
			sendCounter := sendCounter + 1.	
		].
	

		Interpreter enableMetaBehaviorFor: t.
	
		self passAsArg: t. "should trigger meta arg"
		t x: 5. "should trigger meta receive and meta-write".
		self assert: (t x = 5). "should trigger meta-receive and meta-read".
		t testSenderHook. "should trigger 2 meta-sends"
	
		"Interpreter disableMetaBehaviorFor: t."
		Interpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	].
	
	self assert: sendCounter = (2*100).
! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/17/2013 11:41'!
testHookStore

	|t receiveCounter sendCounter ivReadCounter ivWriteCounter argCounter storeCounter |
	
	receiveCounter := 0.
	sendCounter := 0.
	ivReadCounter := 0.
	ivWriteCounter := 0.
	argCounter := 0.
	storeCounter := 0.

	1 to: 100 do: [:i| 	
		
		t := TestObject new.
	

		Interpreter on: Stored for: t do: [
			storeCounter := storeCounter + 1.
		].

		Interpreter enableMetaBehaviorFor: t.
	
		self passAsArg: t. "should trigger meta arg"
		t x: 5. "should trigger meta receive and meta-write".
		self assert: (t x = 5). "should trigger meta-receive and meta-read".
		t testSenderHook. "should trigger 2 meta-sends"
		self store: t.	
		"Interpreter disableMetaBehaviorFor: t."
		Interpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	].
	
	self assert: storeCounter = (1*100).
! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/4/2013 15:46'!
testHookWrite

	|t receiveCounter sendCounter ivReadCounter ivWriteCounter argCounter|
	
	receiveCounter := 0.
	sendCounter := 0.
	ivReadCounter := 0.
	ivWriteCounter := 0.
	argCounter := 0.

	1 to: 100 do: [:i| 	
		
		t := TestObject new.
	

		Interpreter on: IvWrite for: t do: [:reifications :reflectogram | 
			ivWriteCounter := ivWriteCounter + 1.
		].

		Interpreter enableMetaBehaviorFor: t.
	
		self passAsArg: t. "should trigger meta arg"
		t x: 5. "should trigger meta receive and meta-write".
		self assert: (t x = 5). "should trigger meta-receive and meta-read".
		t testSenderHook. "should trigger 2 meta-sends"
	
		"Interpreter disableMetaBehaviorFor: t."
		Interpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	].
	
	self assert: ivWriteCounter = (1*100).
! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/4/2013 15:48'!
testHooks

	|t receiveCounter sendCounter ivReadCounter ivWriteCounter argCounter|
	
	receiveCounter := 0.
	sendCounter := 0.
	ivReadCounter := 0.
	ivWriteCounter := 0.
	argCounter := 0.

	1 to: 100 do: [:i| 	
		
		t := TestObject new.
	
		Interpreter on: MessageReceived for: t do: [:reifications :reflectogram | 
			receiveCounter := receiveCounter + 1.
		].
	
		Interpreter on: MessageSent for: t do: [:reifications :reflectogram | 
			sendCounter := sendCounter + 1.
		].
	
		Interpreter on: IvRead for: t do: [:reifications :reflectogram | 
			ivReadCounter := ivReadCounter + 1.
		].

		Interpreter on: IvWrite for: t do: [:reifications :reflectogram | 
			ivWriteCounter := ivWriteCounter + 1.
		].

		Interpreter on: AsArgument for: t do: [:reifications :reflectogram |
			argCounter := argCounter + 1.
		].

		Interpreter enableMetaBehaviorFor: t.
	
		self passAsArg: t. "should trigger meta arg"
		t x: 5. "should trigger meta receive and meta-write".
		self assert: (t x = 5). "should trigger meta-receive and meta-read".
		t testSenderHook. "should trigger 2 meta-sends"
	
		"Interpreter disableMetaBehaviorFor: t."
		Interpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	].
	
	self assert: receiveCounter = (3*100).
	self assert: sendCounter = (2*100).
	self assert: ivReadCounter = (1*100).
	self assert: ivWriteCounter = (1*100).
	self assert: argCounter = (1*100).! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/4/2013 15:49'!
testHooksWithHalt

	"press proceed for the test to finish"

	|t receiveCounter sendCounter ivReadCounter ivWriteCounter argCounter|
	
	receiveCounter := 0.
	sendCounter := 0.
	ivReadCounter := 0.
	ivWriteCounter := 0.
	argCounter := 0.

		
	t := TestObject new.
	
	Interpreter on: MessageReceived for: t do: [:reifications :reflectogram | 
		self halt.
		receiveCounter := receiveCounter + 1.
	].
	
	Interpreter on: MessageSent for: t do: [:reifications :reflectogram | 
		self halt.
		sendCounter := sendCounter + 1.	
	].
	
	Interpreter on: IvRead for: t do: [:reifications :reflectogram | 
		self halt.
		ivReadCounter := ivReadCounter + 1.
	].

	Interpreter on: IvWrite for: t do: [:reifications :reflectogram | 
		self halt.
		ivWriteCounter := ivWriteCounter + 1.
	].

	Interpreter on: AsArgument for: t do: [:reifications :reflectogram |
		self halt.
		argCounter := argCounter + 1.
	].

		Interpreter enableMetaBehaviorFor: t.
	
		self passAsArg: t. "should trigger meta arg"
		t x: 5. "should trigger meta receive and meta-write".
		self assert: (t x = 5). "should trigger meta-receive and meta-read".
		t testSenderHook. "should trigger 2 meta-sends"
	
		"Interpreter disableMetaBehaviorFor: t."
		Interpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	
	self assert: receiveCounter = 3.
	self assert: sendCounter = 2.
	self assert: ivReadCounter = 1.
	self assert: ivWriteCounter = 1.
	self assert: argCounter = 1.! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/4/2013 15:52'!
testHooksWithOverride

	|t receiveCounter sendCounter ivReadCounter ivWriteCounter argCounter|
	
	receiveCounter := 0.
	sendCounter := 0.
	ivReadCounter := 0.
	ivWriteCounter := 0.
	argCounter := 0.

	1 to: 100 do: [:i| 	
		
		t := TestObject new.
	
		Interpreter on: MessageReceived for: t do: [:reifications :reflectogram | 
			receiveCounter := receiveCounter + 1.
		].
	
		Interpreter on: MessageSent for: t do: [:reifications :reflectogram | 
			sendCounter := sendCounter + 1.
		].
	
		Interpreter on: IvRead for: t do: [:reifications :reflectogram | 
			ivReadCounter := ivReadCounter + 1.
			reflectogram override: true.
			reflectogram returnValue: 666.
			reflectogram enable.
		].

		Interpreter on: IvWrite for: t do: [:reifications :reflectogram | 
			ivWriteCounter := ivWriteCounter + 1.
		].

		Interpreter on: AsArgument for: t do: [:reifications :reflectogram |
			argCounter := argCounter + 1.
		].

		Interpreter enableMetaBehaviorFor: t.
	
		self passAsArg: t. "should trigger meta arg"
		t x: 5. "should trigger meta receive and meta-write".
		self assert: (t x = 666). "should trigger meta-receive and meta-read".
		t testSenderHook. "should trigger 2 meta-sends"
	
		"Interpreter disableMetaBehaviorFor: t."
		Interpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	].
	
	self assert: receiveCounter = (3*100).
	self assert: sendCounter = (2*100).
	self assert: ivReadCounter = (1*100).
	self assert: ivWriteCounter = (1*100).
	self assert: argCounter = (1*100).! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/30/2013 10:55'!
testLevel

	|t receiveCounter level |
	
	receiveCounter := 0.
	
		
	t := TestObject new.
	
	Interpreter on: MessageReceived for: t do: [:reifications :reflectogram | 
			receiveCounter := receiveCounter + 1.
			level := reflectogram processMetaLevel. 
			Transcript show: 'counter: ' , receiveCounter asString , 'level: ' , level asString; cr. 
			(level = 3) ifFalse: [
				reflectogram enable. 
				t message.
				reflectogram disable.
			]
	].

	Interpreter enableMetaBehaviorFor: t.
	t message.
	Interpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	self assert: (receiveCounter = 3).! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/30/2013 11:01'!
testObjectLevel

	|t receiveCounter level |
	
	receiveCounter := 0.
	
		
	t := TestObject new.
	
	Interpreter on: MessageReceived for: t do: [:reifications :reflectogram | 
			receiveCounter := receiveCounter + 1.
			level := reflectogram objectMetaLevel. 
			Transcript show: 'counter: ' , receiveCounter asString , 'level: ' , level asString; cr. 
			(level = 3) ifFalse: [
				reflectogram enable. 
				t message.
				reflectogram disable.
			]
	].

	Interpreter enableMetaBehaviorFor: t.
	t message.
	Interpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	self assert: (receiveCounter = 3).! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/4/2013 15:53'!
testOneShot

	|t receiveCounter sendCounter ivReadCounter ivWriteCounter argCounter|
	
	receiveCounter := 0.

	t := TestObject new.

	Interpreter on: MessageReceived for: t do: [:reifications :reflectogram | 
		receiveCounter := receiveCounter + 1.
		reflectogram override: true.
		reflectogram returnValue: (reflectogram defaultAction).
	].


	Interpreter enableMetaBehaviorFor: t.

	1 to: 100 do: [:i| 	
			
		t class.	
	
	].
	
	"Interpreter disableMetaBehaviorFor: t."
	Interpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	self assert: receiveCounter = 1. "it was one shot"! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/4/2013 15:53'!
testOneShotWithPost

	|t receiveCounter sendCounter ivReadCounter ivWriteCounter argCounter|
	
	receiveCounter := 0.

	t := TestObject new.

	Interpreter on: MessageReceived for: t do: [:reifications :reflectogram | 
		receiveCounter := receiveCounter + 1.
		reflectogram override: true.
		reflectogram returnValue: (reflectogram defaultAction).
		receiveCounter := receiveCounter + 1.
	].


	Interpreter enableMetaBehaviorFor: t.

	1 to: 100 do: [:i| 	
			
		t class.	
	
	].
	
	"Interpreter disableMetaBehaviorFor: t."
	Interpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	self assert: receiveCounter = 2. "it was one shot"! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 9/10/2013 10:28'!
testSetUnset

	|object|
	
	object := Object new.
	Interpreter enableMetaBehaviorFor: object.
	self assert: (Interpreter isMetaBehaviorEnabledFor: object).
	Interpreter disableMetaBehaviorFor: object.
	self assert: (Interpreter isMetaBehaviorEnabledFor: object) not.! !

TestCase subclass: #TestProfileCase
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter-CaseStudy'!

BlockClosure variableSubclass: #BaseAction
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

BlockClosure variableSubclass: #MetaAction
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

Object subclass: #BenchmarkObject
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!BenchmarkObject methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/10/2013 14:41'!
test
	| a b c |
	a := 1.
	b := 2.
	c := a + b.
	^ c
! !

Object subclass: #CallGraph
	instanceVariableNames: 'root helperDict parent current'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter-CaseStudy'!

!CallGraph methodsFor: 'stats' stamp: 'NickPapoulias 3/13/2014 21:00'!
addSelector: aSelector forClass: aClass rcvr: rcvr args: args

	parent := current.
	current := CallGraphNode new.
	current parent: parent.
	parent 
		ifNil: [ self root: current ]
		ifNotNil: [ parent children add: current].
	current selector: aSelector.
	current class: aClass.
	current method: (aClass lookupSelector: aSelector).
	current rcvr: rcvr."rcvr asString." "should be stored as strings"
	current args: args. "(args collect: [ :arg | arg asString ])".
	(helperDict 
		at: current method 
		ifAbsent: [ helperDict at: current method put: OrderedCollection new ]) add: current.
	^ current! !

!CallGraph methodsFor: 'accessing' stamp: 'NickPapoulias 3/13/2014 21:22'!
current

	^ current! !

!CallGraph methodsFor: 'accessing' stamp: 'NickPapoulias 2/27/2014 14:06'!
helperDict
	^ helperDict! !

!CallGraph methodsFor: 'accessing' stamp: 'NickPapoulias 2/27/2014 14:06'!
helperDict: anObject
	helperDict := anObject! !

!CallGraph methodsFor: 'stats' stamp: 'NickPapoulias 2/27/2014 14:39'!
initialize 
	
	self helperDict: Dictionary new.! !

!CallGraph methodsFor: 'stats' stamp: 'NickPapoulias 3/13/2014 21:03'!
return: aNode

	current := aNode.
	parent := aNode parent.! !

!CallGraph methodsFor: 'accessing' stamp: 'NickPapoulias 2/27/2014 14:06'!
root
	^ root! !

!CallGraph methodsFor: 'accessing' stamp: 'NickPapoulias 2/27/2014 14:06'!
root: anObject
	root := anObject! !

!CallGraph methodsFor: 'stats' stamp: 'NickPapoulias 2/27/2014 14:18'!
totalMessageSendsForMethod: aMethod! !

!CallGraph methodsFor: 'stats' stamp: 'NickPapoulias 2/27/2014 14:19'!
totalPrimitivesForMethod: aMethod! !

!CallGraph methodsFor: 'stats' stamp: 'NickPapoulias 2/27/2014 14:18'!
totalTimesCalledForMethod: aMethod! !

Object subclass: #CallGraphNode
	instanceVariableNames: 'rcvr selector args class parent children returnValue'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter-CaseStudy'!

!CallGraphNode methodsFor: 'accessing' stamp: 'NickPapoulias 3/13/2014 20:26'!
args
	^ args! !

!CallGraphNode methodsFor: 'accessing' stamp: 'NickPapoulias 3/15/2014 15:50'!
args: anObject
	args := anObject collect: [:obj | 
		 Interpreter value: [obj asString] forObject: obj.
	]! !

!CallGraphNode methodsFor: 'accessing' stamp: 'NickPapoulias 3/11/2014 18:53'!
children
	^ children! !

!CallGraphNode methodsFor: 'accessing' stamp: 'NickPapoulias 3/11/2014 18:54'!
children: anObject
	children := anObject! !

!CallGraphNode methodsFor: 'accessing' stamp: 'NickPapoulias 3/10/2014 18:34'!
class: anObject
	class := anObject! !

!CallGraphNode methodsFor: 'helper-methods' stamp: 'NickPapoulias 2/27/2014 14:12'!
hasAncestor: aCallGraphNode! !

!CallGraphNode methodsFor: 'helper-methods' stamp: 'NickPapoulias 2/27/2014 14:11'!
hasDescendant: aCallGraphNode! !

!CallGraphNode methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 3/11/2014 18:55'!
initialize 
	
	super initialize.
	children := OrderedCollection new.! !

!CallGraphNode methodsFor: 'helper-methods' stamp: 'NickPapoulias 2/27/2014 14:09'!
isPrimitive! !

!CallGraphNode methodsFor: 'stats' stamp: 'NickPapoulias 2/27/2014 13:51'!
messageSends! !

!CallGraphNode methodsFor: 'accessing' stamp: 'NickPapoulias 2/27/2014 13:32'!
method
	^ method! !

!CallGraphNode methodsFor: 'accessing' stamp: 'NickPapoulias 2/27/2014 13:32'!
method: anObject
	method := anObject! !

!CallGraphNode methodsFor: 'accessing' stamp: 'NickPapoulias 3/10/2014 18:34'!
nodeClass
	^ class! !

!CallGraphNode methodsFor: 'accessing' stamp: 'NickPapoulias 3/11/2014 18:54'!
parent
	^ parent! !

!CallGraphNode methodsFor: 'accessing' stamp: 'NickPapoulias 3/11/2014 18:54'!
parent: anObject
	parent := anObject! !

!CallGraphNode methodsFor: 'stats' stamp: 'NickPapoulias 2/27/2014 13:52'!
primitives! !

!CallGraphNode methodsFor: 'accessing' stamp: 'NickPapoulias 3/15/2014 16:34'!
printString

	^ rcvr , ' ' , selector asString , ' ' , args , ' ^' , returnValue! !

!CallGraphNode methodsFor: 'accessing' stamp: 'NickPapoulias 3/13/2014 20:26'!
rcvr
	^ rcvr! !

!CallGraphNode methodsFor: 'accessing' stamp: 'NickPapoulias 3/15/2014 15:43'!
rcvr: anObject
	rcvr := Interpreter value: [anObject asString] forObject: anObject.! !

!CallGraphNode methodsFor: 'accessing' stamp: 'NickPapoulias 3/15/2014 16:35'!
returnValue
	^ returnValue! !

!CallGraphNode methodsFor: 'accessing' stamp: 'NickPapoulias 3/15/2014 16:35'!
returnValue: anObject
	returnValue := anObject! !

!CallGraphNode methodsFor: 'accessing' stamp: 'NickPapoulias 3/11/2014 18:52'!
selector
	^ selector! !

!CallGraphNode methodsFor: 'accessing' stamp: 'NickPapoulias 3/11/2014 18:52'!
selector: anObject
	selector := anObject! !

Object subclass: #ExecutionTrace
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter-CaseStudy'!

ExecutionTrace class
	instanceVariableNames: 'profileProcess callGraph'!

ExecutionTrace class
	instanceVariableNames: 'profileProcess callGraph'!

!ExecutionTrace class methodsFor: 'tracing' stamp: 'NickPapoulias 3/13/2014 21:48'!
addGraphNode: reflectogram
	|currentRcvr rcvrClass selector|
	currentRcvr := reflectogram reifications receiver.
	rcvrClass := (reflectogram object: currentRcvr perform: #class withArguments: #()).
	selector := reflectogram reifications message selector.
	"Transcript show: currentRcvr; cr."
	"Transcript show: rcvrClass asString , '>>' , selector asString; cr."
	self callGraph
		addSelector: selector
		forClass: rcvrClass
		rcvr: currentRcvr 
		args: reflectogram reifications message arguments.
	[reflectogram on: MessageSent for: currentRcvr do: self callbackSend] on: PrimitiveFailed do: [ 
		reflectogram on: MethodSend for: (rcvrClass lookupSelector: selector) do: self callbackSend.
	].! !

!ExecutionTrace class methodsFor: 'accessing' stamp: 'NickPapoulias 3/11/2014 15:35'!
callGraph
	^ callGraph! !

!ExecutionTrace class methodsFor: 'meta-action' stamp: 'NickPapoulias 3/13/2014 21:21'!
callbackSend
	^ [ :reflectogram | 
	(self inTracingScope: reflectogram) "level control"
		ifTrue: [ 
			|returnNode|
			returnNode := self callGraph current.
			self
				addGraphNode: reflectogram; "set-up / identity control"
				executeNode: reflectogram; "temporal control"
				return: reflectogram fromNode: returnNode] ] "set-up control"! !

!ExecutionTrace class methodsFor: 'tracing' stamp: 'NickPapoulias 3/11/2014 16:50'!
executeNode: reflectogram
	reflectogram
		override: true;
		returnValue: reflectogram defaultAction! !

!ExecutionTrace class methodsFor: 'tracing' stamp: 'NickPapoulais 3/12/2014 13:05'!
inTracingScope: reflectogram
	^ reflectogram reifications process = profileProcess & (reflectogram processMetaLevel = 1)! !

!ExecutionTrace class methodsFor: 'initialization' stamp: 'NickPapoulias 3/11/2014 15:35'!
new
	| instance |
	instance := self basicNew.
	instance initialize.
	callGraph := CallGraph new.
	Interpreter on: MessageSent for: instance do: self callbackSend.
	Interpreter enableMetaBehaviorFor: instance.
	^ instance! !

!ExecutionTrace class methodsFor: 'tracing' stamp: 'NickPapoulias 3/13/2014 21:04'!
return: reflectogram fromNode: aNode	
	|currentRcvr rcvrClass selector|
	currentRcvr := reflectogram reifications receiver.
	rcvrClass := (reflectogram object: currentRcvr perform: #class withArguments: #()).
	selector := reflectogram reifications message selector.
	[reflectogram disableMetaBehaviorFor: currentRcvr] on: PrimitiveFailed do: [
		reflectogram disableMetaBehaviorFor: (rcvrClass lookupSelector: selector).
	].
	self callGraph return: aNode.
	reflectogram enable.! !

!ExecutionTrace class methodsFor: 'initialization' stamp: 'NickPapoulias 3/13/2014 20:27'!
run: aSymbol
	| instance result |
	instance := self new.
	result := #NIL.
	profileProcess := [ 
		(Delay forSeconds: 1) wait.
		self callGraph
			addSelector: aSymbol
			forClass: self
			rcvr: instance 
			args: #().
		result := Interpreter object: instance perform: aSymbol withArguments: #() 
	] fork.
	[ result = #NIL ] whileTrue: [ Processor yield ].
	Interpreter disableMetaBehaviorFor: instance.
	^ {result.
	callGraph}! !

ExecutionTrace subclass: #TracingExample
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter-CaseStudy'!

!TracingExample methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 3/13/2014 20:51'!
fibonacci

	^ 5 benchFib
	! !

ExecutionTrace subclass: #TracingTests
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter-CaseStudy'!

!TracingTests methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 3/13/2014 16:49'!
echo

	self echo2! !

!TracingTests methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 3/13/2014 16:49'!
echo2

	Object new. ! !

!TracingTests methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 3/13/2014 16:49'!
factorial

	^ 10 benchFib
	! !

!TracingTests methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 3/13/2014 16:49'!
fibonacci

	^ 10 benchFib
	! !

!TracingTests methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 3/13/2014 16:49'!
profilingTest

	"testing optimizations"

	| x | 
	x := [  ] size.
	Object class.
	x + 4.
	x = 0 ifTrue: [ x := x+1 ].
	1 to: 10 do: [ :i | Transcript show: 'what ?'; cr.].
	! !

Object subclass: #Interpreter
	instanceVariableNames: ''
	classVariableNames: 'BaseMetaDict'
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!Interpreter class methodsFor: 'initialiazation' stamp: 'NickPapoulias 5/8/2013 15:04'!
baseMetaDict

	^ BaseMetaDict! !

!Interpreter class methodsFor: 'binding' stamp: 'NickPapoulias 5/7/2013 16:12'!
disableMetaBehaviorFor: anObject

	"unbind"
	
	self primitiveSetExperimentalBitOf: anObject to: false. ! !

!Interpreter class methodsFor: 'binding' stamp: 'NickPapoulias 9/2/2013 16:16'!
disableMetaBehaviorForRegistered

	BaseMetaDict keys do: [:key | self disableMetaBehaviorFor: key].! !

!Interpreter class methodsFor: 'binding' stamp: 'NickPapoulias 5/7/2013 16:11'!
enableMetaBehaviorFor: anObject

	"bind"
	
	self primitiveSetExperimentalBitOf: anObject to: true. ! !

!Interpreter class methodsFor: 'binding' stamp: 'NickPapoulias 6/22/2013 16:39'!
enableMetaBehaviorForRegistered

	BaseMetaDict keys do: [:key | self enableMetaBehaviorFor: key].! !

!Interpreter class methodsFor: 'initialiazation' stamp: 'NickPapoulias 3/3/2014 15:09'!
initialize

	super initialize.
	BaseMetaDict := FLLargeIdentityDictionary new.
	Smalltalk specialObjectsArray becomeForward: self newSpecialObjectsArray.
	Smalltalk globals at: #Interpreter put: self. "will put subclass"! !

!Interpreter class methodsFor: 'binding' stamp: 'NickPapoulias 5/7/2013 16:13'!
isMetaBehaviorEnabledFor: anObject


	^ self primitiveGetExperimentalBitOf: anObject! !

!Interpreter class methodsFor: 'logging' stamp: 'NickPapoulias 3/3/2014 13:12'!
log

	^ (Smalltalk specialObjectsArray at: 64) = LoggingInterpreter! !

!Interpreter class methodsFor: 'logging' stamp: 'NickPapoulias 3/3/2014 15:13'!
log: aBoolean

	aBoolean 
		ifTrue: [LoggingInterpreter initialize]
		ifFalse: [LoggingInterpreter superclass initialize].! !

!Interpreter class methodsFor: 'initialiazation' stamp: 'NickPapoulias 3/13/2014 13:25'!
newSpecialObjectsArray
	"Smalltalk recreateSpecialObjectsArray"
	
	"To external package developers:
	**** DO NOT OVERRIDE THIS METHOD.  *****
	If you are writing a plugin and need additional special object(s) for your own use, 
	use addGCRoot() function and use own, separate special objects registry "
	
	"The Special Objects Array is an array of objects used by the Squeak virtual machine.
	 Its contents are critical and accesses to it by the VM are unchecked, so don't even
	 think of playing here unless you know what you are doing."
	
	| newArray |
	newArray := Array new: 79.
	"Nil false and true get used throughout the interpreter"
	newArray at: 1 put: nil.
	newArray at: 2 put: false.
	newArray at: 3 put: true.
	"This association holds the active process (a ProcessScheduler)"
	newArray at: 4 put: (Smalltalk globals associationAt: #Processor).
	"Numerous classes below used for type checking and instantiation"
	newArray at: 5 put: Bitmap.
	newArray at: 6 put: SmallInteger.
	newArray at: 7 put: ByteString.
	newArray at: 8 put: Array.
	newArray at: 9 put: Smalltalk.
	newArray at: 10 put: Float.
	newArray at: 11 put: MethodContext.
	newArray at: 12 put: nil. "Was BlockContext"
	newArray at: 13 put: Point.
	newArray at: 14 put: LargePositiveInteger.
	newArray at: 15 put: Display.
	newArray at: 16 put: Message.
	newArray at: 17 put: CompiledMethod.
	newArray at: 18 put: (Smalltalk specialObjectsArray at: 18).
	"(low space Semaphore)"
	newArray at: 19 put: Semaphore.
	newArray at: 20 put: Character.
	newArray at: 21 put: #doesNotUnderstand:.
	newArray at: 22 put: #cannotReturn:.
	newArray at: 23 put: nil. "This is the process signalling low space."
	"An array of the 32 selectors that are compiled as special bytecodes,
	 paired alternately with the number of arguments each takes."
	newArray at: 24 put: #(	#+ 1 #- 1 #< 1 #> 1 #<= 1 #>= 1 #= 1 #~= 1
							#* 1 #/ 1 #\\ 1 #@ 1 #bitShift: 1 #// 1 #bitAnd: 1 #bitOr: 1
							#at: 1 #at:put: 2 #size 0 #next 0 #nextPut: 1 #atEnd 0 #== 1 nil 0
							nil 1 #value 0 #value: 1 #do: 1 #new 0 #new: 1 #x 0 #y 0 ).
	"An array of the 255 Characters in ascii order.
	 Cog inlines table into machine code at: prim so do not regenerate it."
	newArray at: 25 put: (Smalltalk specialObjectsArray at: 25).
	newArray at: 26 put: #mustBeBoolean.
	newArray at: 27 put: ByteArray.
	newArray at: 28 put: Process.
	"An array of up to 31 classes whose instances will have compact headers"
	newArray at: 29 put: Smalltalk compactClassesArray.
	newArray at: 30 put: (Smalltalk specialObjectsArray at: 30).
	"(delay Semaphore)"
	newArray at: 31 put: (Smalltalk specialObjectsArray at: 31).
	"(user interrupt Semaphore)"
	"Prototype instances that can be copied for fast initialization"
	newArray at: 32 put: (Float new: 2).
	newArray at: 33 put: (LargePositiveInteger new: 4).
	newArray at: 34 put: 0@0.
	newArray at: 35 put: #cannotInterpret:.
	"Note: This must be fixed once we start using context prototypes (yeah, right)"
	"(MethodContext new: CompiledMethod fullFrameSize)."
	newArray at: 36 put: (Smalltalk specialObjectsArray at: 36). "Is the prototype MethodContext (unused by the VM)"
	newArray at: 37 put: BlockClosure.
	"(BlockContext new: CompiledMethod fullFrameSize)."
	newArray at: 38 put: (Smalltalk specialObjectsArray at: 38). "Is the prototype BlockContext (unused by the VM)"
	"array of objects referred to by external code"
	newArray at: 39 put: (Smalltalk specialObjectsArray at: 39).	"preserve external semaphores"
	newArray at: 40 put: nil. "Reserved for Mutex in Cog VMs"
	newArray at: 41 put: nil. "Reserved for a LinkedList instance for overlapped calls in CogMT"
	"finalization Semaphore"
	newArray at: 42 put: ((Smalltalk specialObjectsArray at: 42) ifNil: [Semaphore new]).
	newArray at: 43 put: LargeNegativeInteger.
	"External objects for callout.
	 Note: Written so that one can actually completely remove the FFI."
	newArray at: 44 put: (Smalltalk at: #ExternalAddress ifAbsent: []).
	newArray at: 45 put: (Smalltalk at: #ExternalStructure ifAbsent: []).
	newArray at: 46 put: (Smalltalk at: #ExternalData ifAbsent: []).
	newArray at: 47 put: (Smalltalk at: #ExternalFunction ifAbsent: []).
	newArray at: 48 put: (Smalltalk at: #ExternalLibrary ifAbsent: []).
	newArray at: 49 put: #aboutToReturn:through:.
	newArray at: 50 put: #run:with:in:.
	"51 reserved for immutability message"
	"newArray at: 51 put: #attemptToAssign:withIndex:."
	newArray at: 52 put: #(nil "nil => generic error" #'bad receiver'
							#'bad argument' #'bad index'
							#'bad number of arguments'
							#'inappropriate operation'  #'unsupported operation'
							#'no modification' #'insufficient object memory'
							#'insufficient C memory' #'not found' #'bad method'
							#'internal error in named primitive machinery'
							#'object may move').
	"53 to 55 are for Alien"
	newArray at: 53 put: (Smalltalk at: #Alien ifAbsent: []).
	newArray at: 54 put: #invokeCallback:stack:registers:jmpbuf:.
	newArray at: 55 put: (Smalltalk at: #UnsafeAlien ifAbsent: []).

	"Weak reference finalization"
	newArray at: 56 put: (Smalltalk at: #WeakFinalizationList ifAbsent: []).
	newArray at: 59 put: #metaReceived:.
	newArray at: 60 put: #metaRef.
	newArray at: 61 put: #metaReadAt:.
	newArray at: 62 put: #metaWriteAt:Put:.
	newArray at: 63 put: #metaArgIn:for:.
	newArray at: 64 put: self.
	newArray at: 65 put: #metaSend:from:.
	newArray at: 66 put: self baseMetaDict.
	"optimization"
	newArray at: 67 put: self class>>#object:metaReceived:.
	newArray at: 68 put: self class>>#object:metaSend:withReceiver:.
	newArray at: 69 put: self class>>#object:metaReadAt:.
	newArray at: 70 put: self class>>#object:metaWriteAt:Put:.
	newArray at: 71 put: self class>>#object:metaArgIn:withReceiver:.
	
	newArray at: 72 put: #object:metaReceived:.
	newArray at: 73 put: #object:metaSend:withReceiver:.
	newArray at: 74 put: #object:metaReadAt:.
	newArray at: 75 put: #object:metaWriteAt:Put:.
	newArray at: 76 put: #object:metaArgIn:withReceiver:.
	newArray at: 77 put: #object:metaStoreAt:for:.
	newArray at: 78 put: #object:metaInvoke:withReceiver:.
	newArray at: 79 put: #object:metaMSend:withReceiver:.
	^ newArray.! !

!Interpreter class methodsFor: 'introspection' stamp: 'NickPapoulias 5/5/2013 18:43'!
object: anObject at: anIndex

	<primitive: 73>
	! !

!Interpreter class methodsFor: 'modification' stamp: 'NickPapoulias 3/4/2014 17:24'!
object: anObject at: anIndex put:  anotherObject

	<primitive: 74>
	! !

!Interpreter class methodsFor: 'intercession' stamp: 'NickPapoulias 3/12/2014 14:20'!
object: anObject metaArgIn: aMessage withReceiver: rcvr

	| reflectogram callback |
	"Transcript show: thisContext; show: aMessage ; cr." 	

	callback := (BaseMetaDict at: anObject ifAbsent: [
		BaseMetaDict at: anObject put: LowLevelImplicitMetaObject new.	
	]) argCallback.
	
	callback ifNil: [
		Interpreter enableMetaBehaviorFor: anObject.
		^  Interpreter object: rcvr perform: aMessage selector withArguments: aMessage arguments
	]  ifNotNil:  [
		callback argumentCount = 0 ifTrue: [ 
			callback value.
			Interpreter enableMetaBehaviorFor: anObject.
			^  Interpreter object: rcvr perform: aMessage selector withArguments: aMessage arguments. 
		]. 
		callback argumentCount = 1 ifTrue: [
			callback valueWithArguments: { 				
				reflectogram := Reflectogram 
									for: anObject 
									withDefaultAction:  [Interpreter object: rcvr perform: aMessage selector withArguments: aMessage arguments]
									inContext: thisContext.
			}.
		]ifFalse: [
			callback valueWithArguments: { 
				Reifications with: thisContext . 
				reflectogram := Reflectogram for: anObject withDefaultAction:  [Interpreter object: rcvr perform: aMessage selector withArguments: aMessage arguments]
			}.
		]. 
	].
	
	reflectogram override ifTrue: [
		^ reflectogram returnValue
	] ifFalse: [
		Interpreter enableMetaBehaviorFor: anObject.
		^  Interpreter object: rcvr perform: aMessage selector withArguments: aMessage arguments.				
	]. ! !

!Interpreter class methodsFor: 'intercession' stamp: 'NickPapoulias 3/13/2014 08:57'!
object: anObject metaInvoke: aMessage withReceiver: rcvr

	| reflectogram callback |
	"Transcript show: thisContext; show: aMessage ; cr." 	

	callback := (BaseMetaDict at: anObject ifAbsent: [
		BaseMetaDict at: anObject put: LowLevelImplicitMetaObject new.	
	]) invokeCallback.
	
	callback ifNil: [
		Interpreter enableMetaBehaviorFor: anObject.
		^  Interpreter object: rcvr  perform: aMessage selector withArguments: aMessage arguments.
	]  ifNotNil:  [
		callback argumentCount = 0 ifTrue: [ 
			callback value.
			Interpreter enableMetaBehaviorFor: anObject.
			^   Interpreter object: rcvr  perform: aMessage selector withArguments: aMessage arguments. 
		].
		 callback argumentCount = 1 ifTrue: [
			callback valueWithArguments: { 				
				reflectogram := Reflectogram 
									for: anObject 
									withDefaultAction: [Interpreter object: rcvr  perform: aMessage selector withArguments: aMessage arguments]
									inContext: thisContext.
			}.
		] ifFalse: [
				callback valueWithArguments: { 
				Reifications with: thisContext . 
				reflectogram := Reflectogram for: anObject withDefaultAction:   [Interpreter object: rcvr  perform: aMessage selector withArguments: aMessage arguments]}.
			]. 
	].
	
	reflectogram override ifTrue: [
		^ reflectogram returnValue
	] ifFalse: [
		Interpreter enableMetaBehaviorFor: anObject.
		^ Interpreter object: rcvr  perform: aMessage selector withArguments: aMessage arguments.	
	]. ! !

!Interpreter class methodsFor: 'intercession' stamp: 'NickPapoulias 3/13/2014 13:15'!
object: anObject metaMSend: aMessage withReceiver: rcvr

	| reflectogram callback |
	"Transcript show: thisContext; show: aMessage ; cr." 	

	callback := (BaseMetaDict at: anObject ifAbsent: [
		BaseMetaDict at: anObject put: LowLevelImplicitMetaObject new.	
	]) mSendCallback.
	
	callback ifNil: [
		Interpreter enableMetaBehaviorFor: anObject.
		^  Interpreter object: rcvr  perform: aMessage selector withArguments: aMessage arguments.
	]  ifNotNil:  [
		callback argumentCount = 0 ifTrue: [ 
			callback value.
			Interpreter enableMetaBehaviorFor: anObject.
			^   Interpreter object: rcvr  perform: aMessage selector withArguments: aMessage arguments. 
		].
		 callback argumentCount = 1 ifTrue: [
			callback valueWithArguments: { 				
				reflectogram := Reflectogram 
									for: anObject 
									withDefaultAction: [Interpreter object: rcvr  perform: aMessage selector withArguments: aMessage arguments]
									inContext: thisContext.
			}.
		] ifFalse: [
				callback valueWithArguments: { 
				Reifications with: thisContext . 
				reflectogram := Reflectogram for: anObject withDefaultAction:   [Interpreter object: rcvr  perform: aMessage selector withArguments: aMessage arguments]}.
			]. 
	].
	
	reflectogram override ifTrue: [
		^ reflectogram returnValue
	] ifFalse: [
		Interpreter enableMetaBehaviorFor: anObject.
		^ Interpreter object: rcvr  perform: aMessage selector withArguments: aMessage arguments.	
	]. ! !

!Interpreter class methodsFor: 'intercession' stamp: 'NickPapoulias 3/11/2014 15:53'!
object: anObject metaReadAt: anIndex

	| reflectogram callback |
	"Transcript show: thisContext; show: aMessage ; cr." 	

	callback := (BaseMetaDict at: anObject ifAbsent: [
		BaseMetaDict at: anObject put: LowLevelImplicitMetaObject new.	
	]) readCallback.
	
	callback ifNil: [
		Interpreter enableMetaBehaviorFor: anObject.
		^  Interpreter object: anObject at: anIndex.
	]  ifNotNil:  [
		callback argumentCount = 0 ifTrue: [ 
			callback value.
			Interpreter enableMetaBehaviorFor: anObject.
			^   Interpreter object: anObject at: anIndex. 
		].
		callback argumentCount = 1 ifTrue: [
			callback valueWithArguments: { 				
				reflectogram := Reflectogram 
									for: anObject 
									withDefaultAction: [Interpreter object: anObject at: anIndex]
									inContext: thisContext.
			}.
		] ifFalse: [
				callback valueWithArguments: { 
				Reifications with: thisContext . 
				reflectogram := Reflectogram for: anObject withDefaultAction:   [Interpreter object: anObject at: anIndex]}.
			]. 
	].
	
	reflectogram override ifTrue: [
		^ reflectogram returnValue
	] ifFalse: [
		Interpreter enableMetaBehaviorFor: anObject.
		^ Interpreter object: anObject at: anIndex.				
	]. ! !

!Interpreter class methodsFor: 'intercession' stamp: 'NickPapoulias 3/11/2014 15:54'!
object: anObject metaReceived: aMessage

	| reflectogram callback |
	"Transcript show: thisContext; show: aMessage ; cr." 	

	callback := (BaseMetaDict at: anObject ifAbsent: [
		BaseMetaDict at: anObject put: LowLevelImplicitMetaObject new.	
	]) receiveCallback.
	
	callback ifNil: [
		Interpreter enableMetaBehaviorFor: anObject.
		^ Interpreter object: anObject perform: aMessage selector withArguments: aMessage arguments
	]  ifNotNil:  [
		callback argumentCount = 0 ifTrue: [ 
			callback value.
			Interpreter enableMetaBehaviorFor: anObject.
			^  Interpreter object: anObject perform: aMessage selector withArguments: aMessage arguments. 
		].
		callback argumentCount = 1 ifTrue: [
			callback valueWithArguments: { 				
				reflectogram := Reflectogram 
									for: anObject 
									withDefaultAction: [ Interpreter object: anObject perform: aMessage selector withArguments: aMessage arguments]
									inContext: thisContext.
			}.
		] ifFalse: [
				callback valueWithArguments: { 
				Reifications with: thisContext . 
				reflectogram := Reflectogram for: anObject withDefaultAction:  [ Interpreter object: anObject perform: aMessage selector withArguments: aMessage arguments]}.
			]. 
	].
	
	reflectogram override ifTrue: [
		^ reflectogram returnValue
	] ifFalse: [
		Interpreter enableMetaBehaviorFor: anObject.
		^ Interpreter object: anObject perform: aMessage selector withArguments: aMessage arguments.				
	]. ! !

!Interpreter class methodsFor: 'intercession' stamp: 'NickPapoulias 3/11/2014 16:23'!
object: anObject metaSend: aMessage withReceiver: rcvr

	| reflectogram callback |
	"Transcript show: thisContext; show: aMessage ; cr." 	

	callback := (BaseMetaDict at: anObject ifAbsent: [
		BaseMetaDict at: anObject put: LowLevelImplicitMetaObject new.	
	]) sendCallback.
	
	callback ifNil: [
		Interpreter enableMetaBehaviorFor: anObject.
		^  Interpreter object: rcvr  perform: aMessage selector withArguments: aMessage arguments.
	]  ifNotNil:  [
		callback argumentCount = 0 ifTrue: [ 
			callback value.
			Interpreter enableMetaBehaviorFor: anObject.
			^   Interpreter object: rcvr  perform: aMessage selector withArguments: aMessage arguments. 
		].
		 callback argumentCount = 1 ifTrue: [
			callback valueWithArguments: { 				
				reflectogram := Reflectogram 
									for: anObject 
									withDefaultAction: [Interpreter object: rcvr  perform: aMessage selector withArguments: aMessage arguments]
									inContext: thisContext.
			}.
		] ifFalse: [
				callback valueWithArguments: { 
				Reifications with: thisContext . 
				reflectogram := Reflectogram for: anObject withDefaultAction:   [Interpreter object: rcvr  perform: aMessage selector withArguments: aMessage arguments]}.
			]. 
	].
	
	reflectogram override ifTrue: [
		^ reflectogram returnValue
	] ifFalse: [
		Interpreter enableMetaBehaviorFor: anObject.
		^ Interpreter object: rcvr  perform: aMessage selector withArguments: aMessage arguments.	
	]. ! !

!Interpreter class methodsFor: 'intercession' stamp: 'NickPapoulias 3/11/2014 16:29'!
object: anObject metaStoreAt: anIndex for: anotherObject

	| reflectogram callback |
	"Transcript show: thisContext; show: aMessage ; cr." 	

	callback := (BaseMetaDict at: anObject ifAbsent: [
		BaseMetaDict at: anObject put: LowLevelImplicitMetaObject new.	
	]) storeCallback.
	
	callback ifNil: [
		Interpreter enableMetaBehaviorFor: anObject.
		^   Interpreter object: anotherObject at: anIndex put: anObject
	]  ifNotNil:  [
		callback argumentCount = 0 ifTrue: [ 
			callback value.
			Interpreter enableMetaBehaviorFor: anObject.
			^  Interpreter object: anotherObject at: anIndex put: anObject
		]. 
		callback argumentCount = 1 ifTrue: [
			callback valueWithArguments: { 				
				reflectogram := Reflectogram 
									for: anObject 
									withDefaultAction: [ Interpreter object: anotherObject at: anIndex put: anObject]
									inContext: thisContext.
			}.
		] ifFalse: [
				callback valueWithArguments: { 
				Reifications with: thisContext . 
				reflectogram := Reflectogram for: anObject withDefaultAction:   [ Interpreter object: anotherObject at: anIndex put: anObject]}.
			]. 
	].
	
	reflectogram override ifTrue: [
		^ reflectogram returnValue
	] ifFalse: [
		Interpreter enableMetaBehaviorFor: anObject.
		^  Interpreter object: anotherObject at: anIndex put: anObject.
	]. ! !

!Interpreter class methodsFor: 'intercession' stamp: 'NickPapoulias 3/11/2014 16:29'!
object: anObject metaWriteAt: anIndex Put: aValue

	| reflectogram callback |
	"Transcript show: thisContext; show: aMessage ; cr." 	

	callback := (BaseMetaDict at: anObject ifAbsent: [
		BaseMetaDict at: anObject put: LowLevelImplicitMetaObject new.	
	]) writeCallback.
	
	callback ifNil: [
		Interpreter enableMetaBehaviorFor: anObject.
		^   Interpreter object: anObject at: anIndex put: aValue
	]  ifNotNil:  [
		callback argumentCount = 0 ifTrue: [ 
			callback value.
			Interpreter enableMetaBehaviorFor: anObject.
			^  Interpreter object: anObject at: anIndex put: aValue
		]. 
		callback argumentCount = 1 ifTrue: [
			callback valueWithArguments: { 				
				reflectogram := Reflectogram 
									for: anObject 
									withDefaultAction:  [ Interpreter object: anObject at: anIndex put: aValue]
									inContext: thisContext.
			}.
		] ifFalse: [
				callback valueWithArguments: { 
				Reifications with: thisContext . 
				reflectogram := Reflectogram for: anObject withDefaultAction:   [ Interpreter object: anObject at: anIndex put: aValue]}.
			]. 
	].
	
	reflectogram override ifTrue: [
		^ reflectogram returnValue
	] ifFalse: [
		Interpreter enableMetaBehaviorFor: anObject.
		^  Interpreter object: anObject at: anIndex put: aValue
	]. ! !

!Interpreter class methodsFor: 'execution' stamp: 'NickPapoulias 5/5/2013 18:44'!
object: anObject perform: aSelector withArguments: args 

	<primitive: 84>! !

!Interpreter class methodsFor: 'meta-callback-registration' stamp: 'NickPapoulias 3/9/2014 17:53'!
on: anOperationOrEvent for: anObject do: aCallback

	| implicitMetaObject metaAction |
	
	aCallback class = BlockClosure 
		ifTrue: [ metaAction := MetaAction newFrom: aCallback ]
		ifFalse: [ metaAction := aCallback ].
	
	anOperationOrEvent class = ObjectInteraction ifTrue: [
		anOperationOrEvent registerInteractionCallback: metaAction  for: anObject 
	] ifFalse: [
	
		implicitMetaObject := BaseMetaDict at: anObject ifAbsent: [
			BaseMetaDict at: anObject put: LowLevelImplicitMetaObject new.	
		].

		anOperationOrEvent registerCallback: metaAction to: implicitMetaObject 
	]! !

!Interpreter class methodsFor: 'meta-callback-registration' stamp: 'NickPapoulias 5/8/2013 15:39'!
onArgumentForObject: anObject do: aCallback

	| implicitMetaObject |
	
	implicitMetaObject := BaseMetaDict at: anObject ifAbsent: [
		BaseMetaDict at: anObject put: LowLevelImplicitMetaObject new.	
	].

	implicitMetaObject argCallback: aCallback.
! !

!Interpreter class methodsFor: 'meta-callback-registration' stamp: 'NickPapoulias 5/8/2013 15:39'!
onIvReadForObject: anObject do: aCallback

	| implicitMetaObject |
	
	implicitMetaObject := BaseMetaDict at: anObject ifAbsent: [
		BaseMetaDict at: anObject put: LowLevelImplicitMetaObject new.	
	].

	implicitMetaObject readCallback: aCallback.
! !

!Interpreter class methodsFor: 'meta-callback-registration' stamp: 'NickPapoulias 5/8/2013 15:39'!
onIvWriteForObject: anObject do: aCallback

	| implicitMetaObject |
	
	implicitMetaObject := BaseMetaDict at: anObject ifAbsent: [
		BaseMetaDict at: anObject put: LowLevelImplicitMetaObject new.	
	].

	implicitMetaObject writeCallback: aCallback.
! !

!Interpreter class methodsFor: 'meta-callback-registration' stamp: 'NickPapoulias 5/8/2013 15:39'!
onMessageReceivedForObject: anObject do: aCallback

	| implicitMetaObject |
	
	implicitMetaObject := BaseMetaDict at: anObject ifAbsent: [
		BaseMetaDict at: anObject put: LowLevelImplicitMetaObject new.	
	].

	implicitMetaObject receiveCallback: aCallback.
! !

!Interpreter class methodsFor: 'meta-callback-registration' stamp: 'NickPapoulias 5/8/2013 15:39'!
onMessageSendFromObject: anObject do: aCallback

	| implicitMetaObject |
	
	implicitMetaObject := BaseMetaDict at: anObject ifAbsent: [
		BaseMetaDict at: anObject put: LowLevelImplicitMetaObject new.	
	].

	implicitMetaObject sendCallback: aCallback.
! !

!Interpreter class methodsFor: 'binding' stamp: 'NickPapoulias 5/7/2013 15:47'!
primitiveGetExperimentalBitOf: anObject

	<primitive: 150>
	self primitiveFailed! !

!Interpreter class methodsFor: 'binding' stamp: 'NickPapoulias 5/7/2013 15:22'!
primitiveSetExperimentalBitOf: anObject to: aBoolean

	<primitive: 151>
	self primitiveFailed! !

!Interpreter class methodsFor: 'binding' stamp: 'NickPapoulias 9/2/2013 16:16'!
removeMetaBehaviorForRegistered

	BaseMetaDict keys do: [:key | self removeMetaBehaviorOf: key].! !

!Interpreter class methodsFor: 'binding' stamp: 'NickPapoulias 6/2/2013 16:02'!
removeMetaBehaviorOf: anObject

	"unbind and remove implicit meta-object"
	self disableMetaBehaviorFor: anObject.
	BaseMetaDict removeKey: anObject ifAbsent: [].! !

!Interpreter class methodsFor: 'execution' stamp: 'NickPapoulias 3/15/2014 15:53'!
value: aBlock forObject: anObject 

	| ret enabled|
	[enabled := Interpreter isMetaBehaviorEnabledFor: anObject.
	Interpreter disableMetaBehaviorFor: anObject.
	ret := aBlock value.
	enabled ifTrue: [ Interpreter enableMetaBehaviorFor: anObject ]] on: PrimitiveFailed do: [
		ret := aBlock value.
	].
	^ ret.! !

Interpreter subclass: #LoggingInterpreter
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

LoggingInterpreter class
	instanceVariableNames: 'stdout'!

LoggingInterpreter class
	instanceVariableNames: 'stdout'!

!LoggingInterpreter class methodsFor: 'as yet unclassified' stamp: 'NickPapoulais 3/12/2014 13:04'!
initialize

	"super initialize.
	stdout :=  VTermOutputDriver stdout."! !

!LoggingInterpreter class methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 3/4/2014 10:09'!
object: anObject metaArgIn: aMessage withReceiver: rcvr

	stdout print: 'ARG>>' , anObject asString , '>>' , aMessage asString; lf.
	^ super object: anObject metaArgIn: aMessage withReceiver: rcvr.
	! !

!LoggingInterpreter class methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 3/4/2014 10:09'!
object: anObject metaReadAt: anIndex

	stdout print: 'READ>>' , anObject asString; lf.
	^ super object: anObject metaReadAt: anIndex.
	! !

!LoggingInterpreter class methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 3/10/2014 16:29'!
object: anObject metaReceived: aMessage

	stdout print: 'RECEIVE>>' , anObject asString, '>>' , aMessage asString , '>>' , thisContext sender asString; lf.
	^ super object: anObject metaReceived: aMessage.
! !

!LoggingInterpreter class methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 3/10/2014 09:18'!
object: anObject metaSend: aMessage withReceiver: rcvr

	stdout print: 'SEND>>' , anObject asString , '>>' , rcvr asString , '>>' , aMessage asString; lf.
	^ super object: anObject metaSend: aMessage withReceiver: rcvr.! !

!LoggingInterpreter class methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 3/4/2014 10:10'!
object: anObject metaStoreAt: anIndex for: anotherObject

	stdout print: 'STORE>>' , anObject asString; lf.
	^ super object: anObject metaStoreAt: anIndex for: anotherObject! !

!LoggingInterpreter class methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 3/4/2014 10:10'!
object: anObject metaWriteAt: anIndex Put: aValue

	stdout print: 'WRITE>>' , anObject asString; lf.
	^ super object: anObject metaWriteAt: anIndex Put: aValue! !

Object subclass: #LazyReifications
	instanceVariableNames: 'interpreterContext'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!LazyReifications class methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 3/11/2014 17:56'!
with: anInterpreterContext

	^ self new interpreterContext: anInterpreterContext home! !

!LazyReifications methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 12:10'!
interpreterContext

	^ interpreterContext! !

!LazyReifications methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 12:10'!
interpreterContext: aContext

	interpreterContext := aContext! !

LazyReifications subclass: #Reifications
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!Reifications methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 3/11/2014 17:54'!
context

	^ interpreterContext sender! !

!Reifications methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 12:13'!
index

	^ interpreterContext tempNamed: #anIndex! !

!Reifications methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 11:08'!
message

	^ self interpreterContext tempNamed: #aMessage
! !

!Reifications methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 11:08'!
method

	^ self context method! !

!Reifications methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 3/2/2014 13:59'!
process

	^ Processor activeProcess "caution -- better grab from context"! !

!Reifications methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 12:13'!
receiver

	^ interpreterContext tempNamed: #rcvr
! !

!Reifications methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 11:08'!
sender

	^ self context receiver! !

!Reifications methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 11:08'!
sourceRange

	^ self context debuggerMap 
						rangeForPC: self context pc
						contextIsActiveContext: false 
! !

!Reifications methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 12:13'!
trigger

	^ interpreterContext tempNamed: #anObject! !

!Reifications methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 12:14'!
value

	^ interpreterContext tempNamed: #aValue! !

Object subclass: #LowLevelImplicitMetaObject
	instanceVariableNames: 'receiveCallback sendCallback mSendCallback readCallback writeCallback argCallback storeCallback invokeCallback'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!LowLevelImplicitMetaObject class methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/7/2013 17:57'!
new

	^ self basicNew initialize! !

!LowLevelImplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 5/7/2013 11:48'!
argCallback
	^ argCallback! !

!LowLevelImplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 5/7/2013 11:48'!
argCallback: anObject
	argCallback := anObject! !

!LowLevelImplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 6/2/2013 15:16'!
initialize 

	super initialize.

	"self receiveCallback: [:reifications :override | ].

	self sendCallback: [:reifications :override | ].

	self readCallback: [:reifications :override | ].

	self writeCallback: [:reifications :override | ].
	
	self	 argCallback: [:reifications :override | ].
		"! !

!LowLevelImplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 3/13/2014 08:57'!
invokeCallback
	^ invokeCallback! !

!LowLevelImplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 3/13/2014 08:57'!
invokeCallback: anObject
	invokeCallback := anObject! !

!LowLevelImplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 3/13/2014 13:14'!
mSendCallback
	^ mSendCallback! !

!LowLevelImplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 3/13/2014 13:14'!
mSendCallback: anObject
	mSendCallback := anObject! !

!LowLevelImplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 5/7/2013 11:48'!
readCallback
	^ readCallback! !

!LowLevelImplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 5/7/2013 11:48'!
readCallback: anObject
	readCallback := anObject! !

!LowLevelImplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 5/7/2013 14:58'!
receiveCallback
	^ receiveCallback! !

!LowLevelImplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 5/7/2013 14:58'!
receiveCallback: anObject
	receiveCallback := anObject! !

!LowLevelImplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 5/7/2013 11:48'!
sendCallback
	^ sendCallback! !

!LowLevelImplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 5/7/2013 11:48'!
sendCallback: anObject
	sendCallback := anObject! !

!LowLevelImplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 6/17/2013 11:29'!
storeCallback
	^ storeCallback! !

!LowLevelImplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 6/17/2013 11:29'!
storeCallback: anObject
	storeCallback := anObject! !

!LowLevelImplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 5/7/2013 11:48'!
writeCallback
	^ writeCallback! !

!LowLevelImplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 5/7/2013 11:48'!
writeCallback: anObject
	writeCallback := anObject! !

Object subclass: #MetaBenchmark
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!MetaBenchmark methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/15/2013 14:47'!
benchmark: aBlock times: anInteger
	"synced with Bifrost's bench"
	| runs average deviation |
	aBlock value; value. "this does nothing in our case possible bifrost uses it
	to install methodWrappers the first time this is run ?"
	runs := [ 
		(1 to: anInteger) 
			collect: [ :each | aBlock timeToRun ] ]
		valueUnpreemptively.
	average := runs average asFloat.
	deviation := ((runs 
		detectSum: [ :each | (each - average) * (each - average) ]) 
			/ runs size) sqrt.
	^ Array with: average with: deviation! !

!MetaBenchmark methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/10/2013 14:41'!
message
	"self new message" 
	
	|  object  |
	object := BenchmarkObject new.
	^ self benchmark: [ 1 to: 1000000 do: [ :each | object test ] ] times: 100.! !

!MetaBenchmark methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/5/2013 11:57'!
messageMetaJumpWithHaltOnly
	"self new messageMetaSend" 
	
	|  object meta result |
	object := BenchmarkObject new.
	
	HaltInterpreter initialize.
	 
	HaltInterpreter enableMetaBehaviorFor: object.
	
	self assert: (object test = 3).
	
	result := self
		benchmark: [ 1 to: 1000000 do: [ :each | object test ] ]
		times: 10.
	
	HaltInterpreter removeMetaBehaviorOf: object.
	
	Interpreter initialize.
	
	^ result! !

!MetaBenchmark methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/4/2013 17:10'!
messageMetaSend
	"self new messageMetaSend" 
	
	|  object meta result |
	object := BenchmarkObject new.
	
	Interpreter on: MessageReceived for: object do: [].
	 
	Interpreter enableMetaBehaviorFor: object.
	
	self assert: (object test = 3).
	
	result := self
		benchmark: [ 1 to: 1000000 do: [ :each | object test ] ]
		times: 10.
	
	Interpreter removeMetaBehaviorOf: object.
	
	^ result! !

!MetaBenchmark methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/4/2013 17:51'!
messageMetaSendWithReifications
	"self new messageMetaSend" 
	
	|  object meta result |
	object := BenchmarkObject new.
	
	Interpreter on: MessageReceived for: object do: [:reifications :reflectogram | ].
	 
	Interpreter enableMetaBehaviorFor: object.
	
	self assert: (object test = 3).
	
	result := self
		benchmark: [ 1 to: 1000000 do: [ :each | object test ] ]
		times: 10.
	
	Interpreter removeMetaBehaviorOf: object.
	
	^ result! !

!Object methodsFor: '*MetaInterpreter' stamp: 'NickPapoulias 5/5/2013 18:49'!
experimentalBit
	^ self primitiveGetExperimentalBit ! !

!Object methodsFor: '*MetaInterpreter' stamp: 'NickPapoulias 5/5/2013 18:49'!
experimentalBit: aBoolean
	self primitiveSetExperimentalBitTo: aBoolean! !

!Object methodsFor: '*MetaInterpreter' stamp: 'NickPapoulias 5/8/2013 15:34'!
identityHash
	"Answer a SmallInteger whose value is related to the receiver's identity.
	This method must not be overridden, except by SmallInteger.
	
	Do not override."

	^ (Interpreter object: self perform: #basicIdentityHash withArguments: {}) bitShift: 18! !

!Object methodsFor: '*MetaInterpreter' stamp: 'NickPapoulias 5/5/2013 18:49'!
isExperimentalBitOn
	^ self experimentalBit! !

!Object methodsFor: '*MetaInterpreter' stamp: 'NickPapoulias 5/7/2013 16:07'!
metaArgIn: aMessage for: anObject

	^ Interpreter object: anObject metaArgIn: aMessage withReceiver: self! !

!Object methodsFor: '*MetaInterpreter' stamp: 'NickPapoulias 5/7/2013 16:07'!
metaReadAt: anIndex

	^ Interpreter object: self metaReadAt: anIndex! !

!Object methodsFor: '*MetaInterpreter' stamp: 'NickPapoulias 5/7/2013 16:07'!
metaReceived: aMessage

	^ Interpreter object: self metaReceived: aMessage ! !

!Object methodsFor: '*MetaInterpreter' stamp: 'NickPapoulias 5/5/2013 18:39'!
metaRef

	self halt! !

!Object methodsFor: '*MetaInterpreter' stamp: 'NickPapoulias 5/7/2013 16:07'!
metaSend: aMessage from: anObject

	^ Interpreter object: anObject metaSend: aMessage withReceiver: self.! !

!Object methodsFor: '*MetaInterpreter' stamp: 'NickPapoulias 5/7/2013 16:07'!
metaWriteAt: anIndex Put: anotherObject

	^ Interpreter object: self metaWriteAt: anIndex Put: anotherObject! !

!Object methodsFor: '*MetaInterpreter' stamp: 'NickPapoulias 5/5/2013 18:50'!
primitiveGetExperimentalBit

	<primitive: 150>
	self primitiveFailed! !

!Object methodsFor: '*MetaInterpreter' stamp: 'NickPapoulias 5/5/2013 18:50'!
primitiveSetExperimentalBitTo: aBoolean

	<primitive: 151>
	self primitiveFailed! !

Object subclass: #Operation
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

Operation subclass: #AsArgument
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!AsArgument class methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/4/2013 15:19'!
registerCallback: aCallback to: anImplicitMetaObject

	anImplicitMetaObject argCallback: aCallback ! !

Operation subclass: #AsReturnValue
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

Operation subclass: #CompositeOperation
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

CompositeOperation subclass: #ObjectCreation
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

CompositeOperation subclass: #ObjectInteraction
	instanceVariableNames: 'interactionObject'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!ObjectInteraction class methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/22/2013 18:31'!
with: anotherObject

	^ self new with: anotherObject ! !

!ObjectInteraction methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 9/3/2013 11:51'!
registerInteractionCallback: aCallback for: anObject

	Interpreter on: AsArgument for: anObject do: [:reifications :reflectogram | 
		reifications receiver = interactionObject ifTrue: [aCallback value: reifications value: reflectogram]	
	].
	
	Interpreter on: MessageReceived for: anObject  do: [:reifications :reflectogram | 
		(reifications message arguments includes: interactionObject) ifTrue: [aCallback value: reifications value: reflectogram]	
	].
! !

!ObjectInteraction methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 9/2/2013 14:03'!
with: anotherObject

	interactionObject := anotherObject! !

Operation subclass: #IvRead
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!IvRead class methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/4/2013 15:19'!
registerCallback: aCallback to: anImplicitMetaObject

	anImplicitMetaObject readCallback: aCallback ! !

Operation subclass: #IvWrite
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!IvWrite class methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/4/2013 15:23'!
registerCallback: aCallback to: anImplicitMetaObject

	anImplicitMetaObject writeCallback: aCallback ! !

Operation subclass: #MessageReceived
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!MessageReceived class methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/4/2013 15:24'!
registerCallback: aCallback to: anImplicitMetaObject

	anImplicitMetaObject receiveCallback: aCallback ! !

Operation subclass: #MessageSent
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!MessageSent class methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/4/2013 15:24'!
registerCallback: aCallback to: anImplicitMetaObject

	anImplicitMetaObject sendCallback: aCallback ! !

Operation subclass: #MethodInvoke
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!MethodInvoke class methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 3/13/2014 08:56'!
registerCallback: aCallback to: anImplicitMetaObject

	anImplicitMetaObject invokeCallback: aCallback ! !

Operation subclass: #MethodSend
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!MethodSend class methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 3/13/2014 13:13'!
registerCallback: aCallback to: anImplicitMetaObject

	anImplicitMetaObject mSendCallback: aCallback ! !

!Operation class methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/4/2013 15:15'!
registerCallback: aCallback to: anImplicitMetaObject

	self subclassResponsibility ! !

Operation subclass: #Referenced
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

Operation subclass: #Stored
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!Stored class methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/17/2013 11:19'!
registerCallback: aCallback to: anImplicitMetaObject

	anImplicitMetaObject storeCallback: aCallback ! !

Object subclass: #Reflectogram
	instanceVariableNames: 'trigger returnValue defaultAction override context reifications'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!Reflectogram class methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 12:18'!
for: anObject withDefaultAction: aBlock

	^ self basicNew initializeWithTrigger: anObject andAction: aBlock! !

!Reflectogram class methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 3/11/2014 15:41'!
for: anObject withDefaultAction: aBlock inContext: aContext

	^ self basicNew 
		initializeWithTrigger: anObject 
		andAction: aBlock
		inContext: aContext! !

!Reflectogram methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 3/15/2014 15:38'!
cascade: aBlock

	| ret enabled |
	enabled := Interpreter isMetaBehaviorEnabledFor: trigger.
	Interpreter enableMetaBehaviorFor: trigger.
	ret := aBlock value.
	enabled ifFalse: [Interpreter disableMetaBehaviorFor: trigger].
	^ ret.! !

!Reflectogram methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 3/10/2014 18:07'!
defaultAction

	^ (BaseAction newFrom: defaultAction) value! !

!Reflectogram methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 12:19'!
disable 

	Interpreter disableMetaBehaviorFor: trigger ! !

!Reflectogram methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 3/11/2014 17:16'!
disableMetaBehaviorFor: anObject

	"unbind"
	
	Interpreter disableMetaBehaviorFor: anObject.	! !

!Reflectogram methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 14:56'!
enable

	Interpreter enableMetaBehaviorFor: trigger ! !

!Reflectogram methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 14:45'!
initializeWithTrigger: anObject andAction: aBlock

	trigger := anObject.
	defaultAction := aBlock.
	override := false.
	! !

!Reflectogram methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 3/11/2014 15:41'!
initializeWithTrigger: anObject andAction: aBlock inContext: aContext

	trigger := anObject.
	defaultAction := aBlock.
	override := false.
	context := aContext
	! !

!Reflectogram methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 12:19'!
isEnabled 

	^ Interpreter isMetaBehaviorEnabledFor: trigger ! !

!Reflectogram methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/30/2013 10:33'!
object: anObject at: anIndex

	<primitive: 73>! !

!Reflectogram methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/30/2013 10:33'!
object: anObject at: anIndex put: anotherObject

	<primitive: 74>
	! !

!Reflectogram methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/30/2013 10:34'!
object: anObject perform: aSelector withArguments: args 

	<primitive: 84>! !

!Reflectogram methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 3/10/2014 18:12'!
objectMetaLevel

	|fullStack level|
	
	fullStack := thisContext stackOfSize: 100000.
	level := 0.

	fullStack do: [:ctx | "count the number of value: messages to MetaActions"
		(ctx closure class = MetaAction) ifTrue: [
			(ctx sender arguments at: 1) = trigger ifTrue: [level := level + 1]
		].
		(ctx closure class = BaseAction) ifTrue: [
			(ctx home arguments at: 1) = trigger ifTrue: [level := level - 1]
		] 
	].

	^ level.
	! !

!Reflectogram methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 3/2/2014 14:26'!
on: anOperationOrEvent for: anObject do: aCallback

	Interpreter on: anOperationOrEvent for: anObject do: aCallback.
	Interpreter enableMetaBehaviorFor: anObject.! !

!Reflectogram methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 14:44'!
override

	^ override ! !

!Reflectogram methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 14:43'!
override: aBoolean

	override := aBoolean ! !

!Reflectogram methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 3/10/2014 18:07'!
processMetaLevel

	|fullStack level|
	
	fullStack := thisContext stackOfSize: 100000.
	level := 0.
	
	fullStack do: [:ctx | "count the number of value: messages to MetaActions"
		ctx closure class = MetaAction ifTrue: [level := level + 1].
		ctx closure class = BaseAction ifTrue: [level := level - 1].
	].

	^ level.
	! !

!Reflectogram methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 3/11/2014 15:43'!
reifications

	^ reifications ifNil: [ reifications :=  Reifications with: context]! !

!Reflectogram methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/22/2013 18:12'!
remove

	Interpreter removeMetaBehaviorOf: trigger.! !

!Reflectogram methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 12:19'!
returnValue

	^ returnValue ! !

!Reflectogram methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 12:19'!
returnValue: aValue

	returnValue := aValue! !

!Reflectogram methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 3/15/2014 15:36'!
value: aBlock

	| ret enabled|
	enabled := Interpreter isMetaBehaviorEnabledFor: trigger.
	Interpreter disableMetaBehaviorFor: trigger.
	ret := aBlock value.
	enabled ifTrue: [ Interpreter enableMetaBehaviorFor: trigger ].
	^ ret.! !

Object subclass: #TestObject
	instanceVariableNames: 'x'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!TestObject methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 10/30/2013 10:47'!
message! !

!TestObject methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/11/2013 16:04'!
testInteraction: anObject 

	anObject class.! !

!TestObject methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/11/2013 15:58'!
testPassingArgument: anObject ! !

!TestObject methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/9/2013 19:38'!
testSenderHook

	Transcript show: 'this message should be caught'; cr.! !

!TestObject methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/8/2013 16:01'!
x

	^ x! !

!TestObject methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/5/2013 19:02'!
x: anObject 

	x := anObject ! !

Object subclass: #TracingUI
	instanceVariableNames: 'callgraph roView'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter-CaseStudy'!

!TracingUI class methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 3/14/2014 17:06'!
newWith: aCallGraph 

	^ self new 
		callgraph: (aCallGraph at: 2);
		render! !

!TracingUI methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 3/14/2014 15:02'!
callgraph

	^ callgraph! !

!TracingUI methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 3/14/2014 15:04'!
callgraph: aCallGraph

	callgraph := aCallGraph! !

!TracingUI methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 3/15/2014 14:57'!
decorateEdge: anEdge

	anEdge + (ROLine new add: ROArrow new offset: 0.5).
	^ anEdge! !

!TracingUI methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 3/15/2014 14:56'!
decorateElement: anElement

	anElement width: 100.
	anElement height: 50.
	anElement +  (ROEllipse + ROLabel).
	anElement @ RODraggable.
	^ anElement! !

!TracingUI methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 3/15/2014 14:17'!
render
	
	roView := ROView new.
	self render: callgraph root.
	ROTreeLayout new on: roView elements.
	roView open.! !

!TracingUI methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 3/15/2014 14:51'!
render: aCallGraphNode
	|parentElement|
	parentElement := (ROElement on: aCallGraphNode).
	self decorateElement: parentElement.
	roView add: parentElement.
	aCallGraphNode children do: [:child |
		roView add: 
			(self decorateEdge: 
				(ROEdge lineFrom: parentElement to: (self render: child))).
	 ].
	^ parentElement! !
Interpreter initialize!
LoggingInterpreter initialize!
